@@ -61,7 +61,7 @@ pub trait ContentProviderImplExt: ObjectSubclass {
 
     fn parent_write_mime_type_async<
         Q: IsA<gio::Cancellable>,
-        R: FnOnce(Result<(), glib::Error>) + Send + 'static,
+        R: FnOnce(Result<(), glib::Error>) + 'static,
     >(
         &self,
         provider: &Self::Type,
@@ -163,7 +163,7 @@ impl<T: ContentProviderImpl> ContentProviderImplExt for T {
 
     fn parent_write_mime_type_async<
         Q: IsA<gio::Cancellable>,
-        R: FnOnce(Result<(), glib::Error>) + Send + 'static,
+        R: FnOnce(Result<(), glib::Error>) + 'static,
     >(
         &self,
         provider: &Self::Type,
@@ -174,6 +174,16 @@ impl<T: ContentProviderImpl> ContentProviderImplExt for T {
         callback: R,
     ) {
         unsafe {
+            let main_context = glib::MainContext::ref_thread_default();
+            let is_main_context_owner = main_context.is_owner();
+            let has_acquired_main_context = (!is_main_context_owner)
+                .then(|| main_context.acquire().ok())
+                .flatten();
+            assert!(
+                is_main_context_owner || has_acquired_main_context.is_some(),
+                "Async operations only allowed if the thread is owning the MainContext"
+            );
+
             let data = T::type_data();
             let parent_class = data.as_ref().parent_class() as *mut ffi::GdkContentProviderClass;
             let f = (*parent_class)
@@ -183,10 +193,11 @@ impl<T: ContentProviderImpl> ContentProviderImplExt for T {
                 .write_mime_type_finish
                 .expect("no parent \"write_mime_type_finish\" implementation");
 
-            let user_data: Box<(R, _)> = Box::new((callback, finish));
+            let user_data: Box<(glib::thread_guard::ThreadGuard<R>, _)> =
+                Box::new((glib::thread_guard::ThreadGuard::new(callback), finish));
 
             unsafe extern "C" fn parent_write_mime_type_async_trampoline<
-                R: FnOnce(Result<(), glib::Error>) + Send + 'static,
+                R: FnOnce(Result<(), glib::Error>) + 'static,
             >(
                 source_object_ptr: *mut glib::gobject_ffi::GObject,
                 res: *mut gio::ffi::GAsyncResult,
@@ -194,7 +205,7 @@ impl<T: ContentProviderImpl> ContentProviderImplExt for T {
             ) {
                 let mut error = std::ptr::null_mut();
                 let cb: Box<(
-                    R,
+                    glib::thread_guard::ThreadGuard<R>,
                     fn(
                         *mut ffi::GdkContentProvider,
                         *mut gio::ffi::GAsyncResult,
@@ -207,7 +218,8 @@ impl<T: ContentProviderImpl> ContentProviderImplExt for T {
                 } else {
                     Err(from_glib_full(error))
                 };
-                cb.0(result);
+                let cb = cb.0.into_inner();
+                cb(result);
             }
 
             let cancellable = cancellable.map(|p| p.as_ref());
@@ -239,7 +251,7 @@ impl<T: ContentProviderImpl> ContentProviderImplExt for T {
         Box::pin(gio::GioFuture::new(
             provider,
             move |obj, cancellable, send| {
-                let imp = Self::from_instance(obj);
+                let imp = obj.imp();
 
                 imp.parent_write_mime_type_async(
                     &imp.instance(),
